#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%%%%%%%%%%%%%%%%%% Data augmentation example for GOAL with PIRLS (GOALi) %%%%%%%%%%%%%%%%%%%%%%% 
#%%%%%% This R code is based on OAL R code proposed by Shortreed and Ertefaie (2017) %%%%%%%%%%%%
#%%%%%% and the lqa package proposed by Ulbricht (2010) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

## written for R version 3.6.2 (Author: Ismaila Bald)

## install and load the "MASS" package
install.packages("MASS")
library(MASS) # version 3.3.1

## procedure to install and load the "lqa" package from GitHub in two steps
# first install and load the "devtools" package 
install.packages("devtools")
library(devtools)

# now install and load "lqa" package from GitHub
install_github("cran/lqa")  
library(lqa)

rm(list=ls())  # to remove all objects from the current workspace
ls()           # to check object list
set.seed(2015) # to create repeatable data sets 

## set information for simulating covariates
mean_x = 0 
sig_x = 1

## correlation between covariates
rho = 0.75

## number of Monte Carlo (MC) simulations
S = 1000

## sample size
n = naug = 200

## total number of predictors
p = paug = 100

## pC, pI and pP are number of confounders, predictors of exposure and predictors of outcome, respectively
pC = pI = pP = 2

## pS number of spurious covariates
pS = p - (pC+pI+pP)

## list of variables
var.list = c(paste("Xc",1:pC,sep=""),paste("Xp",1:pP,sep=""),paste("Xi",1:pI,sep=""),paste("Xs",1:pS,sep=""))

## set strength of relationship between covariates and outcome
beta_v = c(0.6, 0.6, 0.6, 0.6, 0, 0, rep(0,p-6))

## set strength of relationship between covariates and treatment
alpha_v = c(1, 1, 0, 0, 1, 1, rep(0,p-6))
names(beta_v) = names(alpha_v) = var.list

## set true average treatment effect
bA = 0

## set vector of possible lambda's to try
lambda_vec = c(-10, -5, -2, -1, -0.75, -0.5, -0.25, 0.25, 0.49)
names(lambda_vec) = as.character(lambda_vec)

## lambda_n (n)^(gamma/2 - 1) = n^(gamma_convergence_factor)
gamma_convergence_factor = 2

## get the gamma value for each value in the lambda vector that corresponds to convergence factor
gamma_vals = 2*(gamma_convergence_factor - lambda_vec + 1)
names(gamma_vals) = names(lambda_vec)

## define some functions for generating data, ATE estimates, and the wAMD,
expit = function(x){ 
  pr = ( exp(x) / (1+exp(x)) ) 
  return(pr)
}
ATE_est = function(fY,fw,fA){
  t_ATE = fY*fw
  tt_ATE = ( ( sum(t_ATE[fA==1]) / sum(fw[fA==1]) ) - ( sum(t_ATE[fA==0]) /  sum(fw[fA==0]) ) )
  return(tt_ATE) 
}
create_weights = function(fp,fA,fw){
  fw = (fp)^(-1)
  fw[fA==0] = (1 - fp[fA==0])^(-1)
  return(fw)
}
wAMD_function = function(DataM,varlist,trt.var,wgt,beta){
  trt = untrt = diff_vec = rep(NA,length(beta)) 
  names(trt) = names(untrt) = names(diff_vec) = varlist
  for(jj in 1:length(varlist)){ 
    this.var = paste("w",varlist[jj],sep="") 
    DataM[,this.var] = DataM[,varlist[jj]] * DataM[,wgt] 
    trt[jj] = sum( DataM[DataM[,trt.var]==1, this.var ]) / sum(DataM[DataM[,trt.var]==1, wgt]) 
    untrt[jj] = sum(DataM[DataM[,trt.var]==0, this.var]) / sum(DataM[DataM[,trt.var]==0, wgt]) 
    diff_vec[jj] = abs( trt[jj] - untrt[jj] ) 
  } 
  wdiff_vec = diff_vec * abs(beta) 
  wAMD = c( sum(wdiff_vec))
  ret = list( diff_vec = diff_vec, wdiff_vec = wdiff_vec, wAMD = wAMD )
  return(ret) 
}

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%%%%% modified lqa.update2 from the lqa package proposed by Ulbricht (2010) %%%%%%%%%%%%%%%%%%%%%%%%
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lqa.upd <-
  function (x, y, family = NULL, penalty = NULL, intercept = TRUE, weights = rep (1, nobs), control = lqa.control (), initial.beta, mustart, eta.new, gamma1 = 1, ...)
  {
    gamma <- gamma1
    
    if (is.null (family))
      stop ("lqa.update: family not specified")
    
    if (is.null (penalty))
      stop ("lqa.update: penalty not specified")
    
    if (!is.null (dim (y)))
      stop ("lqa.update: y must be a vector")
    
    x <- as.matrix (x)
    converged <- FALSE
    n.iter <- control$max.steps
    eps <- control$conv.eps
    c1 <- control$c1
    
    stop.at <- n.iter
    pcol <- ncol (x)
    nobs <- nrow (x)
    converged <- FALSE
    beta.mat <- matrix (0, nrow = n.iter, ncol = pcol)    # to store the coefficient updates
    
    if (missing (initial.beta))
      initial.beta <- rep (0.01, pcol)
    else
      eta.new <- drop (x %*% initial.beta)
    
    if (missing (mustart))
    {
      etastart <- drop (x %*% initial.beta)
      eval (family$initialize)
    }
    
    if (missing (eta.new))
      eta.new <- family$linkfun (mustart)    # predictor
    
    for (i in 1 : n.iter)
    {
      beta.mat[i,] <- initial.beta  
      mu.new <- family$linkinv (eta.new)      # fitted values
      d.new <- family$mu.eta (eta.new)        # derivative of response function
      v.new <- family$variance (mu.new)       # variance function of the response
      weights <- as.vector(d.new / sqrt (v.new)) # decomposed elements (^0.5) of weight matrix W, see GLM notation
      weights <- c(weights[1:naug],rep(1,paug))
      x.star <- weights * x   
      y.tilde.star <- as.vector(weights * (eta.new  + (y - mu.new) / d.new)) 
      y.tilde.star <- c( y.tilde.star[1:naug],rep(0,paug))  
      A.lambda <- get.Amat (initial.beta = initial.beta, penalty = penalty, intercept = intercept, c1 = c1, x = x) 
      p.imat.new <- crossprod (x.star) + A.lambda       # penalized information matrix
      
      chol.pimat.new <- chol (p.imat.new)               # applying cholesky decomposition for matrix inversion
      inv.pimat.new <- chol2inv (chol.pimat.new)        # inverted penalized information matrix
      beta.new <- gamma * drop (inv.pimat.new %*% t (x.star) %*% y.tilde.star) + (1 - gamma) * beta.mat[i,]  # computes the next iterate of the beta vector
      
      if ((sum (abs (beta.new - initial.beta)) / sum (abs (initial.beta)) <= eps))    # check convergence condition
      {
        converged <- TRUE
        stop.at <- i
        if (i < n.iter)
          break
      } 
      else
      {
        initial.beta <- beta.new    # update beta vector
        eta.new <- drop (x %*% beta.new)      
      }
    }
    
    Hatmat <- x.star %*% inv.pimat.new %*% t (x.star)
    tr.H <- sum (diag (Hatmat))
    dev.m <- sum (family$dev.resids (y, mu.new, weights))
    
    aic.vec <- dev.m + 2 * tr.H
    bic.vec <- dev.m + log (nobs) * tr.H
    
    if (!converged & (stop.at == n.iter))
      cat ("lqa.update with ", penalty$penalty, ": convergence warning! (lambda = ", penalty$lambda, ")\n")
    
    fit <- list (coefficients = beta.new, beta.mat = beta.mat[1 : stop.at,], tr.H = tr.H, fitted.values = mu.new, family = family, Amat = A.lambda, converged = converged, stop.at = stop.at, m.stop = stop.at, linear.predictors = eta.new, weights = weights^2, p.imat = p.imat.new, inv.pimat = inv.pimat.new, x.star = x.star, v.new = v.new)
  }

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%%%%%%% modified lqa.default from the lqa package proposed by Ulbricht (2010) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

lqa.def <- function (x, y, family = binomial (), penalty = NULL, weights = rep (1, nobs), 
                     start = NULL, etastart = NULL, mustart = NULL, offset = rep (0, nobs), 
                     control = lqa.control (), intercept = TRUE, standardize = TRUE,method="lqa.update2")
{
  call <- match.call ()
  
  ## check for exponential family and link function:
  ## -----------------------------------------------
  
  if (is.character (family)) 
    family <- get (family, mode = "function", envir = parent.frame ())
  
  if (is.function (family)) 
    family <- family ()
  
  if (is.null (family$family)) 
  {
    print (family)
    stop ("'family' not recognized")
  }
  
  ## check for quadratic penalty:
  ## ---------------------------- 
  
  if (! (method == "nng.update"))
  {
    
    if (is.null (penalty))
      stop ("penalty not specified \n")
    
    if (is.character (penalty)) 
      penalty <- get (penalty, mode = "function", envir = parent.frame ())
    
    if (is.function (penalty)) 
      penalty <- penalty ()
    
    if (is.null (penalty$penalty))
    {
      print (penalty)
      stop ("'penalty' not recognized")
    }
  }
  
  ## check for existence of 'method':
  ## --------------------------------
  
  if (is.null (method))
    stop ("method not specified")
  
  ## check for column of ones in x if intercept == TRUE:
  ## ---------------------------------------------------
  
  if (intercept & (var (x[,1]) > control$var.eps))
    x <- cbind (1, x)
  
  
  ## standardization:
  ## ----------------
  
  x <- as.matrix (x)
  xnames <- dimnames (x)[[2L]]
  
  ynames <- if (is.matrix (y)) 
    rownames(y)
  else 
    names(y)
  
  nobs <- nrow (x)   
  nvars <- ncol (x)    # number of coefficients in the predictor (including an intercept, if present)
  ones <- rep (1, nobs)
  mean.x <- drop (ones %*% x) / nobs       # computes the vector of means
  
  if (intercept)    # if an intercept is included in the model its corresponding element of mean.x is set to zero
    mean.x[1] <- 0  # (such that x[,1] is not getting centered (and also not standardized later on ...))
  
  x.std <- scale (x, mean.x, FALSE)   # centers the regressor matrix
  
  norm.x <- if (standardize)
  {
    norm.x <- sqrt (drop (ones %*% (x.std^2)))   # computes the euclidean norm of the regressors
    nosignal <- apply (x, 2, var) < control$var.eps
    if (any (nosignal))    # modify norm.x for variables with too small a variance (e.g. the intercept)
      norm.x[nosignal] <- 1
    
    norm.x
  }
  else
    rep (1, nvars)
  
  x.std <- scale (x.std, FALSE, norm.x)  # standardizes the centered regressor matrix
  
  ## call and get the (estimation) method: 
  ## -------------------------------------
  
  fit=lqa.upd(x = x.std, y = y, family = family, penalty = penalty, intercept = intercept, 
              control = control)
  
  ## back-transformation of estimated coefficients:
  ## ----------------------------------------------
  
  coef <- fit$coefficients
  
  if (intercept)
  {
    coef[1] <- coef[1] - sum (mean.x[-1] * coef[-1] / norm.x[-1])
    coef[-1] <- coef[-1] / norm.x[-1]
  }
  else
    coef <- coef / norm.x
  
  ## computation of some important statistics:
  ## -----------------------------------------
  
  # remark: the predictors are identical no matter whether we use standardized or unstandardized values, hence all statistics
  #         based on the predictor eta are also equal
  
  eta <- drop (x %*% coef)
  mu <- family$linkinv (eta)
  mu.eta.val <- family$mu.eta (eta)
  wt <- sqrt ((weights * mu.eta.val^2) / family$variance (mu))
  dev <- sum (family$dev.resids (y, mu, weights))
  wtdmu <- sum (weights * y) / sum (weights)
  nulldev <- sum (family$dev.resids (y, wtdmu, weights))
  n.ok <- nobs - sum (weights == 0)
  nulldf <- n.ok - as.integer (intercept)
  residuals <- (y - mu) / mu.eta.val
  
  xnames <- colnames (x)
  ynames <- names (y)
  names (residuals) <- names (mu) <- names (eta) <- names (weights) <- names (wt) <- ynames
  names (coef) <- xnames
  
  Amat <- fit$Amat
  Amat <- t (norm.x * t (norm.x * Amat))   # must be (quadratically) transformed in order to cope with the transformed parameter space
  
  if (is.null (fit$tr.H))
    stop ("quadpen.fit: Element 'tr.H' has not been returned from 'method'")
  
  model.aic <- dev + 2 * fit$tr.H
  model.bic <- dev + log (nobs) * fit$tr.H
  resdf <- n.ok - fit$tr.H
  
  dispersion <- ifelse (!((family$family == "binomial") | (family$family == "poisson")), sum ((y - mu)^2 / family$variance (mu)) / (nobs - fit$tr.H), 1)
  
  fit <- list (coefficients = coef, residuals = residuals, fitted.values = mu, family = family, penalty = penalty, 
               linear.predictors = eta, deviance = dev, aic = model.aic, bic = model.bic, null.deviance = nulldev, n.iter = fit$stop.at, 
               best.iter = fit$m.stop, weights = wt, prior.weights = weights, df.null = nulldf, df.residual = resdf, converged = fit$converged, mean.x = mean.x, 
               norm.x = norm.x, Amat = Amat, method = method, rank = fit$tr.H, x = x, y = y, fit.obj = fit, call = call, dispersion = dispersion)
  
  class (fit) <- c ("lqa", "glm", "lm")
  fit
}

## want to save ATE and propensity score coefficients for the optimal lambda2 value
WGOAL_i=GOAL_i=rep(NA,S)
mati=matrix(NA,S,p)
n.p=n+p

for(s in 1:S){
  Data=NULL
  ## simulate data
  Sigma_x = matrix(rho*sig_x^2,nrow=length(var.list),ncol=length(var.list)) 
  diag(Sigma_x) = sig_x^2
  Mean_x = rep(mean_x,length(var.list))
  Data = as.data.frame(mvrnorm(n = n,mu=Mean_x,Sigma = Sigma_x,empirical = FALSE))
  names(Data) = var.list
  gA_x = rowSums(Data[,var.list]*matrix(alpha_v,nrow=n,ncol=length(var.list),byrow=TRUE))
  pA = expit(gA_x)
  Data$A = as.numeric(runif(n=length(pA)) < pA) # simulate A 
  gY_xA = rowSums(Data[,var.list]*matrix(beta_v,nrow=n,ncol=length(var.list),byrow=TRUE))   
  Data$Y = gY_xA + rnorm(n=n,sd=sig_x)
  Data$Y = Data$Y + Data$A*bA
  
  # normalize covariates to have mean 0 and standard deviation 1
  temp.mean = colMeans(Data[,var.list])
  Temp.mean = matrix(temp.mean,ncol=length(var.list),nrow=nrow(Data),byrow=TRUE)
  Data[,var.list] = Data[,var.list] - Temp.mean
  temp.sd = apply(Data[var.list],FUN=sd,MARGIN=2)
  Temp.sd = matrix(temp.sd,ncol=length(var.list),nrow=nrow(Data),byrow=TRUE)
  Data[var.list] = Data[,var.list] / Temp.sd
  rm(list=c("temp.mean","Temp.mean","temp.sd","Temp.sd"))
  w.full.form = formula(paste("A~",paste(var.list,collapse="+")))
  
  # estimate outcome model
  y.form = formula(paste("Y~A+",paste(var.list,collapse="+")))
  lm.Y = lm(y.form,data=Data)
  betaXY = coef(lm.Y)[var.list] 
  
  ## want to save ATE, wAMD and propensity score coefficients for each lambda value
  ATE = wAMD_vec = rep(NA, length(lambda_vec))
  names(ATE) = names(wAMD_vec) = names(lambda_vec)
  coeff_XA = as.data.frame(matrix(NA,nrow=length(var.list),ncol=length(lambda_vec)))
  names(coeff_XA) = names(lambda_vec)
  rownames(coeff_XA) = var.list
  
  ## set the possible lambda2 value (taken from Zou and Hastie (2005))
  S_lam=c(0,10^c(-2,-1.5,-1,-0.75,-0.5,-0.25,0,0.25,0.5,1))
  ## want to save ATE, wAMD and propensity score coefficients for each lambda2 value
  WM_P=M_P=S_wamd=rep(NA,length(S_lam))
  M_mat=matrix(NA,length(S_lam),p)
  for (m in 1:length(S_lam)) {
    ## create augmented A and X 
    lambda2=S_lam[m]
    I=diag(1,p,p)
    Ip=sqrt(lambda2)*I
    Anp=c(Data$A,rep(0,p))
    Xnp=matrix(0,n+p,p)
    X=Data[,var.list]
    for (j in 1:p){
      Xnp[,j]=c(X[,j],Ip[,j])
    }
    newData=as.data.frame(Xnp)
    names(newData)=var.list
    newData$A=Anp
    
    ## want to save ATE, wAMD and propensity score coefficients for each lambda value
    ATE_try=ATE = wAMD_vec =coeff_XA=NULL;
    ATE_try=ATE = wAMD_vec = rep(NA, length(lambda_vec))
    names(ATE) = names(wAMD_vec) = names(lambda_vec)
    coeff_XA = as.data.frame(matrix(NA,nrow=length(var.list),ncol=length(lambda_vec)))
    names(coeff_XA) = names(lambda_vec)
    rownames(coeff_XA) = var.list
    
    ## run GOAL based on PIRLS 
    # weight model with all possible covariates included, this is passed into lasso function
    for( lil in names(lambda_vec) ){
      il = lambda_vec[lil]
      ig = gamma_vals[lil]
      
      # create the outcome adaptive lasso penalty with coefficient specific weights determined by outcome model
      oal_pen = adaptive.lasso(lambda=n.p^(il),al.weights = abs(betaXY)^(-ig) )
      # run outcome-adaptive lasso model with appropriate penalty
      X=as.matrix(newData[var.list]);y=as.vector(newData$A);
      lq22=lqa.def(x=X, y=y, family = binomial,penalty = oal_pen)
      
      # generate propensity score for ATE
      Data[,paste("f.pA",lil,sep="")]=expit(as.matrix(cbind(rep(1,n),Data[var.list]))%*%as.matrix((1+lambda2)*coef(lq22)))
      # create inverse probability of treatment weights for ATE
      Data[,paste("w",lil,sep="")] = create_weights(fp=Data[,paste("f.pA",lil,sep="")],fA=Data$A)
      # save propensity score coef
      coeff_XA[var.list,lil] = (1+lambda2)*coef(lq22)[var.list]
      # estimate weighted absolute mean different over all covariates using this lambda to generate weights
      wAMD_vec[lil] = wAMD_function(DataM=Data,varlist=var.list,trt.var="A",
                                    wgt=paste("w",lil,sep=""),beta=betaXY)$wAMD
      
      # save ATE estimate for this lambda value
      ATE[lil] = ATE_est(fY=Data$Y,fw=Data[,paste("w",lil,sep="")],fA=Data$A)
    } # close loop through lambda values
    
    # print out wAMD for all the lambda values evaluated
    wAMD_vec
    # find the lambda value that creates the smallest wAMD
    tt = which.min(wAMD_vec)
    # print out ATE corresponding to smallest wAMD value
    ATE[tt][[1]]
    # save the coefficients for the propensity score that corresponds to smallest wAMD value 
    GOAL.i.c=coeff_XA[,tt]
    # check which covariates are selected
    M_mat[m,]=ifelse(abs(coeff_XA[,tt])> 10^(-8),1,0)
    # save the ATE corresponding to smallest wAMD value
    GOAL_est1=ATE[tt][[1]]
    GOAL.i.ate=GOAL_est1
    M_P[m]=GOAL.i.ate
    # save the smallest wAMD value
    WM_P[m]=wAMD_vec[tt][[1]]
    
  } 
  
  ## final ATE, WAMD and propensity score coefficients
  # find the optimal lambda2 value that creates the smallest wAMD
  ptt= which.min(WM_P)
  # save the ATE corresponding the optimal lambda2 value 
  GOAL_i[s]=M_P[ptt]
  # save the coefficients corresponding the optimal lambda2 value 
  mati[s,]=M_mat[ptt,]
  # save wAMD value corresponding the optimal lambda2 value 
  WGOAL_i[s]=WM_P[ptt]
  
  ## print the current iteration
  print(s)
}

## mean estimate from the S replications of ATE.
Moy_GOAL_i=mean(GOAL_i)

## GOALi: Bias, Variance, MSE
Bias_Gi=Moy_GOAL_i-bA;
Var_Gi=(1/(S-1))*sum((GOAL_i-rep(Moy_GOAL_i,S))^2);
MSE_Gi=(1/S)*sum((GOAL_i-rep(bA,S))^2);
Resume_Gi=c(Bias_Gi,Var_Gi,MSE_Gi)

## print tabulated results
Results=rbind(GOALi=Resume_Gi)
colnames(Results)=c("Bias","Variance","MSE")
Results=as.data.frame(round(Results,2))
View(Results)

## box plot
boxplot(GOAL_i, main="GOALi",ylim=c(-1.5,1.5))
lines(0:p,rep(bA,p+1),lty=2, col="red",lwd=3)

## proportion of times covariate selected
plot(1:p,(1/S)*colSums(mati),ylim=c(0,1), type="l",lty=1, col="blue", axes=TRUE, ann=FALSE,lwd=3)
# label the x and y axes 
title(xlab="Covariate index")
title(ylab="Proportion of times covariate selected ")
legend("topright",
       "GOALi",
       col="blue", 
       lwd=3, 
       lty=1 
)


