#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%%%%%%%%%%%%%%%%%% Data augmentation example for naive GOAL (GOALn) %%%%%%%%%%%%%%%%%%%%%%%%%%%% 
#%%%%%% This R code is based on OAL R code proposed by Shortreed and Ertefaie 2017 %%%%%%%%%%%%%%
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

## written for R version 3.6.2 (Author: Ismaila Bald)

## install and load the "MASS" package
install.packages("MASS")
library(MASS) # version 3.3.1

## procedure to install and load the "lqa" package from GitHub in two steps

# first install and load the "devtools" package 
install.packages("devtools") 
library(devtools)

# now install and load "lqa" package from GitHub
install_github("cran/lqa")  
library(lqa)

rm(list=ls())  # to remove all objects from the current workspace
ls()           # to check object list
set.seed(2015) # to create repeatable data sets 

## set information for simulating covariates
mean_x = 0 
sig_x = 1

## correlation between covariates
rho = 0.75

## number Monte carlo (MC) simulations
S = 1000

## sample size
n = 200

## total number of predictors
p = 100

## pC, pI and pP are number of confounders, predictors of exposure and predictors of outcome, respectively
pC = pI = pP = 2

## pS number of spurious covariates
pS = p - (pC+pI+pP)

## list of variables
var.list = c(paste("Xc",1:pC,sep=""),paste("Xp",1:pP,sep=""),paste("Xi",1:pI,sep=""),paste("Xs",1:pS,sep=""))

## set strength of relationship between covariates and outcome
beta_v =  c(0.6, 0.6, 0.6, 0.6, 0, 0, rep(0,p-6))

## set strength of relationship between covariates and treatment
alpha_v = c(1, 1, 0, 0, 1, 1,  rep(0,p-6))
names(beta_v) = names(alpha_v) = var.list

## set true average treatment effect
bA = 0

## set vector of possible lambda's to try
lambda_vec = c(-10, -5, -2, -1, -0.75, -0.5, -0.25, 0.25, 0.49)
names(lambda_vec) = as.character(lambda_vec)

## lambda_n (n)^(gamma/2 - 1) = n^(gamma_convergence_factor)
gamma_convergence_factor = 2

## get the gamma value for each value in the lambda vector that corresponds to convergence factor
gamma_vals = 2*(gamma_convergence_factor - lambda_vec + 1)
names(gamma_vals) = names(lambda_vec)

## define some functions for generating data, ATE estimates, and the wAMD,
expit = function(x){ 
  pr = ( exp(x) / (1+exp(x)) ) 
  return(pr)
}
ATE_est = function(fY,fw,fA){
  t_ATE = fY*fw
  tt_ATE = ( ( sum(t_ATE[fA==1]) / sum(fw[fA==1]) ) - ( sum(t_ATE[fA==0]) /  sum(fw[fA==0]) ) )
  return(tt_ATE) 
}
create_weights = function(fp,fA,fw){
  fw = (fp)^(-1)
  fw[fA==0] = (1 - fp[fA==0])^(-1)
  return(fw)
}
wAMD_function = function(DataM,varlist,trt.var,wgt,beta){
  trt = untrt = diff_vec = rep(NA,length(beta)) 
  names(trt) = names(untrt) = names(diff_vec) = varlist
  for(jj in 1:length(varlist)){ 
    this.var = paste("w",varlist[jj],sep="") 
    DataM[,this.var] = DataM[,varlist[jj]] * DataM[,wgt] 
    trt[jj] = sum( DataM[DataM[,trt.var]==1, this.var ]) / sum(DataM[DataM[,trt.var]==1, wgt]) 
    untrt[jj] = sum(DataM[DataM[,trt.var]==0, this.var]) / sum(DataM[DataM[,trt.var]==0, wgt]) 
    diff_vec[jj] = abs( trt[jj] - untrt[jj] ) 
  } 
  wdiff_vec = diff_vec * abs(beta) 
  wAMD = c( sum(wdiff_vec))
  ret = list( diff_vec = diff_vec, wdiff_vec = wdiff_vec, wAMD = wAMD )
  return(ret) 
}

## want to save ATE and propensity score coefficients for the optimal lambda2 value
WGOAL_l=GOAL_l=rep(NA,S)
mat=matrix(NA,S,p)
n.p=n+p

## MC simulation
for(s in 1: S){
  Data=NULL
  ## simulate data
  Sigma_x = matrix(rho*sig_x^2,nrow=length(var.list),ncol=length(var.list)) 
  diag(Sigma_x) = sig_x^2
  Mean_x = rep(mean_x,length(var.list))
  Data = as.data.frame(mvrnorm(n = n,mu=Mean_x,Sigma = Sigma_x,empirical = FALSE))
  names(Data) = var.list
  gA_x = rowSums(Data[,var.list]*matrix(alpha_v,nrow=n,ncol=length(var.list),byrow=TRUE))
  pA = expit(gA_x)
  Data$A = as.numeric(runif(n=length(pA)) < pA) # simulate A 
  gY_xA = rowSums(Data[,var.list]*matrix(beta_v,nrow=n,ncol=length(var.list),byrow=TRUE))   
  Data$Y = gY_xA + rnorm(n=n,sd=sig_x)
  Data$Y = Data$Y + Data$A*bA
  
  ## normalize covariates to have mean 0 and standard deviation 1
  temp.mean = colMeans(Data[,var.list])
  Temp.mean = matrix(temp.mean,ncol=length(var.list),nrow=nrow(Data),byrow=TRUE)
  Data[,var.list] = Data[,var.list] - Temp.mean
  temp.sd = apply(Data[var.list],FUN=sd,MARGIN=2)
  Temp.sd = matrix(temp.sd,ncol=length(var.list),nrow=nrow(Data),byrow=TRUE)
  Data[var.list] = Data[,var.list] / Temp.sd
  rm(list=c("temp.mean","Temp.mean","temp.sd","Temp.sd"))
  w.full.form = formula(paste("A~",paste(var.list,collapse="+")))
  
  ## estimate outcome model
  y.form = formula(paste("Y~A+",paste(var.list,collapse="+")))
  lm.Y = lm(y.form,data=Data)
  betaXY = coef(lm.Y)[var.list] 
  
  ## want to save ATE, wAMD and propensity score coefficients for each lambda value
  ATE = wAMD_vec =coeff_XA=NULL
  ATE = wAMD_vec = rep(NA, length(lambda_vec))
  names(ATE) = names(wAMD_vec) = names(lambda_vec)
  coeff_XA = as.data.frame(matrix(NA,nrow=length(var.list),ncol=length(lambda_vec)))
  names(coeff_XA) = names(lambda_vec)
  rownames(coeff_XA) = var.list
  
  ## set the possible lambda2 value (taken from Zou and Hastie (2005))
  S_lam=c(0,10^c(-2,-1.5,-1,-0.75,-0.5,-0.25,0,0.25,0.5,1))
  ## want to save ATE, wAMD and propensity score coefficients for each lambda2 value
  WM_N=M_N=S_wamd=rep(NA,length(S_lam))
  M_mat=matrix(NA,length(S_lam),p)
  for (m in 1:length(S_lam)) {
    ## create augmented A and X 
    lambda2=S_lam[m]
    I=diag(1,p,p)
    Ip=sqrt(lambda2)*I
    Anp=c(Data$A,rep(0,p))
    Xnp=matrix(0,n+p,p)
    X=Data[,var.list]
    for (j in 1:p){
      Xnp[,j]=c(X[,j],Ip[,j])
    }
    newData=as.data.frame(Xnp)
    names(newData)=var.list
    newData$A=Anp
    
    ## want to save ATE, wAMD and propensity score coefficients for each lambda value
    ATE_try=ATE = wAMD_vec =coeff_XA=NULL;
    ATE_try=ATE = wAMD_vec = rep(NA, length(lambda_vec))
    names(ATE) = names(wAMD_vec) = names(lambda_vec)
    coeff_XA = as.data.frame(matrix(NA,nrow=length(var.list),ncol=length(lambda_vec)))
    names(coeff_XA) = names(lambda_vec)
    rownames(coeff_XA) = var.list
    
    ## run GOAL with lqa using augmented data
    # weight model with all possible covariates included, this is passed into lasso function
    w.full.form = formula(paste("A~",paste(var.list,collapse="+")))
    for( lil in names(lambda_vec) ){
      il = lambda_vec[lil]
      ig = gamma_vals[lil]
      
      # create the outcome adaptive lasso penalty with coefficient specific weights determined by outcome model
      oal_pen = adaptive.lasso(lambda=n.p^(il),al.weights = abs(betaXY)^(-ig) )
      # run outcome-adaptive lasso model with appropriate penalty
      logit_oal = lqa( w.full.form, data=newData, penalty=oal_pen, family=binomial(logit) )
      # save propensity score coefficients
      coeff_XA[var.list,lil] = (1+lambda2)*coef(logit_oal)[var.list]
      # generate propensity score 
      Data[,paste("f.pA",lil,sep="")]=expit(as.matrix(cbind(rep(1,n),Data[var.list]))%*%as.matrix((1+lambda2)*coef(logit_oal)))
      # create inverse probability of treatment weights for ATE
      Data[,paste("w",lil,sep="")] = create_weights(fp=Data[,paste("f.pA",lil,sep="")],fA=Data$A)
      # estimate weighted absolute mean different over all covaraites using this lambda to generate weights
      wAMD_vec[lil] = wAMD_function(DataM=Data,varlist=var.list,trt.var="A",
                                    wgt=paste("w",lil,sep=""),beta=betaXY)$wAMD
      # save ATE estimate for this lambda value
      ATE[lil] = ATE_est(fY=Data$Y,fw=Data[,paste("w",lil,sep="")],fA=Data$A)
      
    } # close loop through lambda values
    
    # print out wAMD for all the lambda values evaluated
    wAMD_vec
    # find the lambda value that creates the smallest wAMD
    tt = which.min(wAMD_vec)
    # print out ATE corresponding to smallest wAMD value
    ATE[tt][[1]]
    # save the coefficients for the propensity score that corresponds to smallest wAMD value 
    GOAL.lqa.c=coeff_XA[,tt]
    # check which covariates are selected
    M_mat[m,]=ifelse(abs(coeff_XA[,tt])> 10^(-8),1,0)
    # save the ATE corresponding to smallest wAMD value
    GOAL_est1=ATE[tt][[1]]
    GOAL.lqa.ate=GOAL_est1
    M_N[m]=GOAL.lqa.ate
    # save the smallest wAMD value
    WM_N[m]=wAMD_vec[tt][[1]]
  } 
  
  ## final ATE, wAMD and propensity score coefficients
  
  # find the optimal lambda2 value that creates the smallest wAMD
  ntt= which.min(WM_N)
  # save the ATE corresponding the optimal lambda2 value
  GOAL_l[s]=M_N[ntt]
  # save the propensity score coefficients corresponding the optimal lambda2 value 
  mat[s,]=M_mat[ntt,]
  # save wAMD value corresponding the optimal lambda2 value
  WGOAL_l[s]=WM_N[ntt]
  
  ## print the current iteration
  print(s)
}

## mean estimate from the S replications of the ATE
Moy_GOAL_l=mean(GOAL_l)

## GOALn: Bias,Variance and MSE 
Bias_Gl=Moy_GOAL_l-bA;
Var_Gl=(1/(S-1))*sum((GOAL_l-rep(Moy_GOAL_l,S))^2);
MSE_Gl=(1/S)*sum((GOAL_l-rep(bA,S))^2);
Resume_Gl=c(Bias_Gl,Var_Gl,MSE_Gl)

## print tabulated results
Results=rbind(GOALn=Resume_Gl)
colnames(Results)=c("Bias","Variance","MSE")
Results=as.data.frame(round(Results,2))
View(Results)

## box plot
boxplot(GOAL_l, names=c("GOALn"),ylim=c(-1.5,1.5))
lines(0:p,rep(bA,p+1),lty=2, col="red",lwd=3)

## variable selection
plot(1:p,(1/S)*colSums(mat),ylim=c(0,1), type="l",lty=1, col="red", axes=TRUE, ann=FALSE,lwd=3)
# label the x and y axes 
title(xlab="Covariate index")
title(ylab="Proportion of times covariate selected ")
legend("topright", 
       c("GOALn"), 
       col="red", 
       lwd=3, 
       lty=1
)


